q_create_function = '''
mutation CreateFunction($name: String!, $sourceCode: String, $sourceFile: String, $sourceLine: Int, $language: String, $feature: String!, $functionSetId: ID){
    createFunction(input: {name: $name, sourceCode: $sourceCode, sourceFile: $sourceFile, sourceLine: $sourceLine, language: $language, feature: $feature, functionSetId: $functionSetId}){
        function {
            id
        }
    }
}
'''

q_query_function = '''
query QueryFunction($funcId: ID!){
    function(id: $funcId){
        id
        name
        sourceCode
        sourceFile
        sourceLine
        language
    }
}
'''

q_create_function_set = '''
mutation CreateFunctionSet($functionIds: [ID!]){
    createFunctionSet(input: {functionIds: $functionIds}){
        functionSet{
            id
        }
    }
}
'''

q_query_function_set = '''
query QueryFuncitonSet($funcSetId: ID!){
    functionSet(id: $funcSetId){
        id
        functions {
            id
        }
    }
}
'''

q_search_func_similarity = '''
query SearchFuncSimilarity($funcId: ID!, $setId: [ID!], $topk: Int!){
    function(id: $funcId){
        similarity(functionSetIds: $setId, topK: $topk){
            score
            function{
                id
                name
                sourceCode
                language
            }
        }
    }
}
'''



# coding: utf-8
import os
import json
import idaapi
import idautils
import binaryai as bai
from binaryai import BinaryAIException


class BinaryAIManager:
    Default = {
        'token': '',
        'url': '',
        'funcset': '',
        'topk': 10,
        'minsize': 3,
        'threshold': 0.8
    }

    def __init__(self):
        self.name = "BinaryAI"
        cfg_dir = os.path.join(idaapi.get_user_idadir(), idaapi.CFG_SUBDIR)
        os.makedirs(cfg_dir) if not os.path.exists(cfg_dir) else None
        self.cfg = Config(os.path.join(cfg_dir, "{}.cfg".format(bai.__name__)), BinaryAIManager.Default)
        self._client = None
        self.cview = None

    @property
    def client(self):
        if self._client is None:
            if not self.cfg['token']:
                self.cfg['token'] = idaapi.ask_str("", 0, "{} Token:".format(self.name))
            assert self.cfg['token']
            if self.cfg['url']:
                self._client = bai.client.Client(self.cfg['token'], self.cfg['url'])
            else:
                self._client = bai.client.Client(self.cfg['token'])
        return self._client

    @property
    def funcset_id(self):
        if not self.cfg['funcset']:
            self.cfg['funcset'] = bai.function.create_function_set(self.client)
        assert self.cfg['funcset']
        return self.cfg['funcset']

    def upload_function(self, ea, funcset_id):
        func_feat = bai.ida.get_func_feature(ea)
        func_name = idaapi.get_func_name(ea)
        hf = idaapi.hexrays_failure_t()
        cfunc = idaapi.decompile(ea, hf)
        if func_feat and func_name:
            func_id = bai.function.upload_function(
                self.client, func_name, func_feat, source_code=str(cfunc), funcset_id=funcset_id)
            return func_id

    def retrieve_function(self, ea, topk, funcset_ids):
        func_id = self.upload_function(ea, None)
        if func_id:
            try:
                targets = bai.function.search_sim_funcs(self.client, func_id, funcset_ids=funcset_ids, topk=topk)
            except BinaryAIException as e:
                print(e)
                if e.code == "INVALID_ARGUMENT_TOPK_EXCEED_CAPACITY":
                    return e.data['function']['similarity']
            else:
                return targets

    def retrieve_selected_functions(self, funcs):
        btn_type = idaapi.ask_yn(idaapi.ASKBTN_NO, "AUTOHIDE REGISTRY\nSearch in private functionset?")
        if btn_type == idaapi.ASKBTN_CANCEL:
            return
        funcset_ids = [self.funcset_id] if btn_type == idaapi.ASKBTN_YES else None
        for ea in funcs:
            pfn = idaapi.get_func(ea)
            func_name = idaapi.get_func_name(ea)
            if idaapi.FlowChart(pfn).size < self.cfg['minsize']:
                continue
            targets = self.retrieve_function(ea, topk=1, funcset_ids=funcset_ids)
            if targets is None:
                print("[{}] {} is skipped because get function feature error".format(self.name, func_name))
                continue
            func = targets[0]
            if func['score'] < self.cfg['threshold']:
                continue
            pfn.flags |= idaapi.FUNC_LUMINA
            idaapi.update_func(pfn)
            comment = SourceCodeViewer.source_code_comment(func_name, func)
            idaapi.set_func_cmt(pfn, comment, 0)

    def upload_selected_functions(self, funcs):
        succ, skip, fail = 0, 0, 0
        for ea in funcs:
            pfn = idaapi.get_func(ea)
            if idaapi.FlowChart(pfn).size < self.cfg['minsize']:
                skip += 1
                continue
            func_id = self.upload_function(ea, self.funcset_id)
            func_name = idaapi.get_func_name(ea)
            if not func_id:
                print("[{}] {} is skipped because upload error".format(self.name, func_name))
                fail += 1
                continue
            succ += 1
        print("[{}] {} functions successfully uploaded, {} functions failed, {} functions skipped".format(
            self.name, succ, fail, skip))

    def binaryai_calllback(self, __):
        print("[{}] v{}".format(self.name, bai.__version__))

    def retrieve_function_callback(self, __, ea=None):
        btn_type = idaapi.ask_yn(idaapi.ASKBTN_NO, "AUTOHIDE REGISTRY\nSearch in private functionset?")
        if btn_type == idaapi.ASKBTN_CANCEL:
            return
        funcset_ids = [self.funcset_id] if btn_type == idaapi.ASKBTN_YES else None
        func_ea = idaapi.get_screen_ea() if ea is None else ea
        func_name = idaapi.get_func_name(func_ea)
        targets = self.retrieve_function(func_ea, self.cfg['topk'], funcset_ids)
        if targets is None:
            print("[{}] {} is skipped because get function feature error".format(self.name, func_name))
            return

        if not (self.cview and self.cview.is_alive()):
            self.cview = SourceCodeViewer(self.name)
            # CDVF_STATUSBAR 0x04, keep the status bar in the custom viewer
            idaapi.set_code_viewer_is_source(idaapi.create_code_viewer(self.cview.GetWidget(), 0x4))
        self.cview.set_user_data(func_name, targets)

        widget = idaapi.get_current_widget()
        if idaapi.get_widget_title(widget) == self.name:
            return

        if idaapi.get_widget_type(widget) != idaapi.BWN_PSEUDOCODE:
            widget = idaapi.open_pseudocode(func_ea, 1).toplevel
        pseudo_title = idaapi.get_widget_title(widget)

        idaapi.display_widget(self.cview.GetWidget(), idaapi.PluginForm.WOPN_DP_TAB | idaapi.PluginForm.WOPN_RESTORE)
        idaapi.set_dock_pos(self.name, pseudo_title, idaapi.DP_RIGHT)
        return

    def retrieve_all_callback(self, __):
        self.retrieve_selected_functions(idautils.Functions())

    def upload_function_callback(self, __, ea=None):
        func_ea = idaapi.get_screen_ea() if ea is None else ea
        func_id = self.upload_function(func_ea, self.funcset_id)
        func_name = idaapi.get_func_name(func_ea)
        if not func_id:
            print("[{}] {} is skipped because upload error".format(self.name, func_name))
        else:
            print("[{}] {} successfully uploaded".format(self.name, func_name))

    def upload_all_callback(self, __):
        self.upload_selected_functions(idautils.Functions())


class Config(dict):
    def __init__(self, path, default):
        self.path = path
        if not os.path.exists(path):
            json.dump(default, open(self.path, 'w'), indent=4)
        self.cfg = json.load(open(path))

    def __getitem__(self, key):
        return self.cfg[key]

    def __setitem__(self, key, val):
        self.cfg[key] = val
        json.dump(self.cfg, open(self.path, 'w'), indent=4)


class SourceCodeViewer(idaapi.simplecustviewer_t):
    @staticmethod
    def source_code_comment(query, func, idx=0):
        return """/*
    query:  {}
    target[{}]: {}
    score:  {}
*/\n""".format(query, idx, func['function']['name'], func['score'])

    @staticmethod
    def source_code_body(func):
        body = func['function']['sourceCode'].split("\n")
        return filter(lambda l: not l.lstrip().startswith('#'), body)

    def __init__(self, title):
        idaapi.simplecustviewer_t.__init__(self)
        self.alive = True
        self.Create(title)

    def is_alive(self):
        return self.alive

    def set_user_data(self, query, targets):
        self.idx = 0
        self.query = query
        self.targets = targets
        self._repaint()

    def _repaint(self):
        self.ClearLines()
        func = self.targets[self.idx]
        for line in self.source_code_comment(self.query, func, self.idx).split("\n"):
            self.AddLine(idaapi.COLSTR(line, idaapi.SCOLOR_RPTCMT))
        for line in self.source_code_body(func):
            self.AddLine(str(line))
        self.Refresh()

    def OnClose(self):
        self.alive = False

    def OnKeydown(self, vkey, shift):
        if shift == 0 and vkey == ord("K"):
            self.idx = (self.idx + len(self.targets) - 1) % len(self.targets)
            self._repaint()
        elif shift == 0 and vkey == ord("J"):
            self.idx = (self.idx + 1) % len(self.targets)
            self._repaint()


class UIManager:
    class UIHooks(idaapi.UI_Hooks):
        def finish_populating_widget_popup(self, widget, popup):
            if idaapi.get_widget_type(widget) == idaapi.BWN_FUNCS:
                idaapi.attach_action_to_popup(widget, popup, "BinaryAI:RetrieveSelected", "BinaryAI/")
                idaapi.attach_action_to_popup(widget, popup, "BinaryAI:UploadSelected", "BinaryAI/")

    class ActionHandler(idaapi.action_handler_t):
        def __init__(self, name, label, shortcut=None, tooltip=None, icon=-1, flags=0):
            idaapi.action_handler_t.__init__(self)
            self.name = name
            self.action_desc = idaapi.action_desc_t(name, label, self, shortcut, tooltip, icon, flags)

        def register_action(self, callback, toolbar_name=None, menupath=None):
            self.callback = callback
            if not idaapi.register_action(self.action_desc):
                return False
            if toolbar_name and not idaapi.attach_action_to_toolbar(toolbar_name, self.name):
                return False
            if menupath and not idaapi.attach_action_to_menu(menupath, self.name, idaapi.SETMENU_APP):
                return False
            return True

        def activate(self, ctx):
            self.callback(ctx)

        def update(self, ctx):
            return idaapi.AST_ENABLE_ALWAYS

    def __init__(self, name):
        self.name = name
        self.mgr = BinaryAIManager()
        self.hooks = UIManager.UIHooks()

    def register_actions(self):
        toolbar_name, menupath = self.name, self.name
        idaapi.create_toolbar(toolbar_name, self.name)
        idaapi.create_menu(menupath, self.name, "Help")

        UIManager.ActionHandler(self.name, self.name).register_action(self.mgr.binaryai_calllback, toolbar_name)
        action = UIManager.ActionHandler("BinaryAI:RetrieveFunction", "Retrieve function", "Ctrl+Shift+d", icon=99)
        action.register_action(self.mgr.retrieve_function_callback, toolbar_name, menupath)
        action = UIManager.ActionHandler("BinaryAI:RetrieveAll", "Retrieve all functions", "", icon=188)
        action.register_action(self.mgr.retrieve_all_callback, toolbar_name, menupath)
        action = UIManager.ActionHandler("BinaryAI:UploadFunction", "Upload function", "", icon=97)
        action.register_action(self.mgr.upload_function_callback, toolbar_name, menupath)
        action = UIManager.ActionHandler("BinaryAI:UploadAll", "Upload all functions", "", icon=88)
        action.register_action(self.mgr.upload_all_callback, toolbar_name, menupath)

        retrieve_action = UIManager.ActionHandler("BinaryAI:RetrieveSelected", "Retrieve")
        upload_action = UIManager.ActionHandler("BinaryAI:UploadSelected", "Upload")
        if retrieve_action.register_action(self.selected_callback) and \
                upload_action.register_action(self.selected_callback):
            self.hooks.hook()
            return True
        return False

    def selected_callback(self, ctx):
        funcs = map(idaapi.getn_func, ctx.chooser_selection)
        funcs = map(lambda func: func.start_ea, funcs)
        if ctx.action == "BinaryAI:RetrieveSelected":
            self.mgr.retrieve_selected_functions(funcs)
        if ctx.action == "BinaryAI:UploadSelected":
            self.mgr.upload_selected_functions(funcs)


class Plugin(idaapi.plugin_t):
    wanted_name = "BinaryAI"
    comment, help, wanted_hotkey = "", "", ""
    flags = idaapi.PLUGIN_FIX | idaapi.PLUGIN_HIDE

    def init(self):
        if idaapi.init_hexrays_plugin():
            mgr = UIManager(Plugin.wanted_name)
            if mgr.register_actions():
                return idaapi.PLUGIN_OK
        return idaapi.PLUGIN_SKIP

    def run(self, ctx):
        return

    def term(self):
        return


def PLUGIN_ENTRY():
    return Plugin()



class BinaryAIException(Exception):
    def __init__(self, code, msg, data, raw):
        super().__init__()
        self._code = code
        self._msg = msg
        self._data = data
        self._raw = raw

    @property
    def code(self):
        return self._code

    @property
    def data(self):
        return self._data

    def __str__(self):
        return "{}: {}".format(self._code, self._msg)
        


from .error import BinaryAIException
try:
    import idaapi
    import idautils
except ImportError:
    raise BinaryAIException("SDK_ERROR", "Failed to import idaapi or idautils", None, None)

import ctypes
import hashlib
import json


M_MAX = 0x49  # first unused opcode
All_STR = dict(map(lambda i: (i.ea, str(i)), idautils.Strings()))


class Graph:
    def __init__(self):
        self.graph = {"directed": True, "graph": {}, "nodes": [], "links": [], "multigraph": False}
        self.nodes_num = 0
        self.edges_num = 0

    def add_node(self, serial, feat):
        self.graph["nodes"].append({"id": serial, "feat": feat})
        self.nodes_num += 1

    def add_edge(self, serial, succ):
        self.graph["links"].append({"source": serial, "target": succ})
        self.edges_num += 1

    def remove_featempty_nodes(self):
        remove_nodes = []
        for node in self.graph["nodes"]:
            if not node["feat"]:
                self.remove_edge_by_node(node["id"])
                remove_nodes.append(node)
        for node in remove_nodes:
            self.graph["nodes"].remove(node)
            self.nodes_num -= 1

    def remove_edge_by_node(self, id):
        remove_edges = []
        for edge in self.graph["links"]:
            if edge["source"] == id or edge["target"] == id:
                remove_edges.append(edge)
        for edge in remove_edges:
            self.graph["links"].remove(edge)
            self.edges_num -= 1

    def have_nodes(self):
        return self.nodes_num > 0

    def have_edges(self):
        return self.edges_num > 0


def get_idb_info():
    info = idaapi.get_inf_structure()
    if info.is_64bit():
        bits = '64'
    elif info.is_32bit():
        bits = '32'
    else:
        assert(False)

    return ''.join([info.procName, bits])


class CtreeFeature(idaapi.ctree_visitor_t):
    def __init__(self, state, expr, num, stri):
        idaapi.ctree_visitor_t.__init__(self, idaapi.CV_FAST)
        self.state = state
        self.expr = expr
        self.num = num
        self.stri = stri

    def visit_expr(self, item):
        self.expr.append(item.op)
        if item.op == idaapi.cot_num:
            self.num.append(ctypes.c_long(item.n._value).value)
        elif item.op == idaapi.cot_obj:
            addr = item.obj_ea
            if addr in All_STR:
                s = All_STR[addr]
                if s is not None:
                    self.stri.append(s)
        elif item.op == idaapi.cot_str:
            self.stri.append(item.string)

        return 0

    def visit_insn(self, item):
        self.state.append(item.op)
        return 0


def parse_minsn(minsn, micro_int, ins=None):
    ins = [] if ins is None else ins
    ins.append(minsn.opcode)
    for op in [minsn.l, minsn.r, minsn.d]:
        if op.t == idaapi.mop_d:
            parse_minsn(op.d, micro_int, ins)
        elif op.t == idaapi.mop_f:
            for arg in op.f.args:
                if arg.t == idaapi.mop_d:
                    parse_minsn(arg.d, micro_int, ins)
                else:
                    ins.append(arg.t + M_MAX)
        else:
            ins.append(op.t + M_MAX)
            if op.t == idaapi.mop_n:
                micro_int.append(ctypes.c_long(op.nnn.value).value)
    return ins


def parse_func(pfn):
    try:
        hf = idaapi.hexrays_failure_t()
        cfunc = idaapi.decompile(pfn.start_ea, hf)
        mbr = idaapi.mba_ranges_t(pfn)
        mba = idaapi.gen_microcode(
            mbr,
            hf,
            None,
            idaapi.DECOMP_NO_WAIT | idaapi.DECOMP_NO_CACHE,
            idaapi.MMAT_GLBOPT3
        )
    except Exception:
        return
    if mba is None:
        return

    G = Graph()
    ctree_state, ctree_expr, ctree_int, ctree_str, micro_int = [], [], [], [], []

    # node level
    for i in range(mba.qty):
        mb = mba.get_mblock(i)
        minsn = mb.head
        blk = []
        while minsn:
            ins = parse_minsn(minsn, micro_int)
            blk.append(ins)
            minsn = minsn.next

        G.add_node(mb.serial, feat=blk)
        for succ in mb.succset:
            G.add_edge(mb.serial, succ)
    G.remove_featempty_nodes()

    if not G.have_nodes():
        return

    # add a fake edge if there is no edge
    if not G.have_edges():
        G.add_edge(G.graph['nodes'][0]['id'], G.graph['nodes'][0]['id'])

    # graph level
    ctree_fea = CtreeFeature(ctree_state, ctree_expr, ctree_int, ctree_str)
    ctree_fea.apply_to(cfunc.body, None)

    G.graph['graph']['c_state'], G.graph['graph']['c_expr'], G.graph['graph']['c_int'], G.graph['graph'][
        'c_str'], G.graph['graph']['m_int'] = ctree_state, ctree_expr, ctree_int, ctree_str, micro_int
    G.graph['graph']['arg_num'] = len(cfunc.argidx)

    func_bytes = b''
    for start, end in idautils.Chunks(pfn.start_ea):
        fb = idaapi.get_bytes(start, end-start)
        func_bytes += fb
    G.graph['graph']['hash'] = hashlib.md5(func_bytes).hexdigest()

    return G.graph


def get_func_feature(ea):
    """
    get function feature by IDA Pro

    Args:
        ea(ea_t): function address

    Returns:
        func_feat(string): function feature
    """
    pfn = idaapi.get_func(ea)
    G = parse_func(pfn)
    if G:
        func_feat = json.dumps(G)
        return func_feat
    return None



import requests
import json
from .error import BinaryAIException


class Client(object):
    '''
    BinaryAI api client

    Attributes:
        token(string): token used for query
        url(string): BinaryAI api endpoint url, default is https://api.binaryai.tencent.com/v1/endpoint
        timeout(int): seconds of timeout, default is 1000
    '''

    def __init__(
        self,
        token,
        url="https://api.binaryai.tencent.com/v1/endpoint",
        timeout=1000,
    ):
        self.url = url
        self.headers = {
            "Content-Type": "application/json",
            "token": token
        }
        self.session = None
        self.GMT_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'
        self.timeout = timeout

    def _gen_gql_data(self, query, var):
        data = {
            "operationName": None,
            "variables": var,
            "query": query
        }
        return data

    def execute(self, query, var):
        '''
        Send request to API server and get response

        Args:
            query(string): graphql query
            var(dict): variables to send

        Returns:
            dict: response data of this request
        '''
        if not self.session:
            self.session = requests.Session()
        data = self._gen_gql_data(query, var)
        response = None
        try:
            response = self.session.post(self.url, data=json.dumps(
                data), headers=self.headers, timeout=self.timeout)
        except Exception as e:
            raise BinaryAIException("SDK_ERROR", "Request failed with exception: {}".format(e), None, None)

        try:
            jdata = json.loads(response.content)
        except Exception:
            raise BinaryAIException("SDK_ERROR", "Invalid response: {}".format(response.content), None, None)

        if "errors" in jdata.keys():
            errors = jdata["errors"][0]
            raise BinaryAIException(errors['extensions']['code'], errors['message'], jdata['data'], jdata)

        if "data" not in jdata.keys():
            raise BinaryAIException("SDK_ERROR", "Invalid response: {}".format(response.content), None, None)

        return jdata["data"]



#!/usr/bin/env python
from binaryai.client import Client
from binaryai.function import query_function, create_function_set, query_function_set, search_sim_funcs
import click


@click.group(invoke_without_command=True)
@click.option('--url', '-u', type=str, help='api url', default="https://api.binaryai.tencent.com/v1/endpoint", show_default=True)
@click.option('--token', '-t', type=str, help='user token')
@click.option('--help', '-h', is_flag=True, help='Show this message and exit.')
@click.option('--version', '-v', is_flag=True, help='Show version')
@click.pass_context
def cli(ctx, url, token, help, version):
    if ctx.invoked_subcommand is None or help:
        if version:
            import binaryai
            click.echo(binaryai.__version__)
            ctx.exit()
        else:
            banner = r'''
 ____  _                           _    ___
| __ )(_)_ __   __ _ _ __ _   _   / \  |_ _|
|  _ \| | '_ \ / _` | '__| | | | / _ \  | |
| |_) | | | | | (_| | |  | |_| |/ ___ \ | |
|____/|_|_| |_|\__,_|_|   \__, /_/   \_\___|
                          |___/
        '''
            click.echo(banner)
            click.echo(ctx.get_help())
            ctx.exit()
    client = Client(token, url)
    ctx.obj = client


@cli.command('query_function', short_help='get function info by given id')
@click.option('--funcid', '-f', help='function id', required=True)
@click.pass_context
def QueryFunction(ctx, funcid):
    client = ctx.obj
    result = query_function(client, funcid)
    click.echo(result)


@cli.command('create_funcset', short_help='create a new function set and add functions if needed')
@click.option('--funcid', '-f', multiple=True, type=str, help='function ids, set multi ids with -i id1 -i id2')
@click.pass_context
def CreateFuncSet(ctx, funcid):
    client = ctx.obj
    result = create_function_set(client, list(funcid))
    click.echo({'funcsetid': result})


@cli.command('query_funcset', short_help='get function set info by id')
@click.option('--funcset', '-s', help='funcset id', type=str, required=True)
@click.pass_context
def QueryFuncSet(ctx, funcset):
    client = ctx.obj
    result = query_function_set(client, funcset)
    click.echo(result)


@cli.command('search_funcs', short_help='search top similar functions of the query')
@click.option('--funcid', '-f', help='function id', type=str, required=True)
@click.option('--funcset', '-s', multiple=True, type=str, help='funcset ids, set multi ids with -s id1 -s id2')
@click.option('--topk', '-k', type=int, default=10, show_default=True, help='return first topk results')
@click.pass_context
def SearchFuncs(ctx, funcid, funcset, topk):
    client = ctx.obj
    result = search_sim_funcs(client, funcid, list(funcset), topk)
    click.echo(result)


def main():
    cli()



from binaryai import client, function
from .error import BinaryAIException


def _get_version(default='version not found'):
    try:
        from pkg_resources import DistributionNotFound, get_distribution
    except ImportError:
        return default
    else:
        try:
            return get_distribution(__name__).version
        except DistributionNotFound:
            return default


__version__ = _get_version()

__all__ = [
    'client',
    'function',
    'BinaryAIException'
]

try:
    import idaapi
    from binaryai import ida
except ImportError:
    pass
else:
    __all__.append('ida')



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup, find_packages

setup(
    name='binaryai',
    author='binaryai',
    author_email='binaryai@tencent.com',
    packages=find_packages(),
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    install_requires=['requests', 'Click'],
    url="https://github.com/binaryai/sdk",
    license='GPLv3',
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://binaryai.tencent.com",
        "Source": "https://github.com/binaryai/sdk",
    },
    python_requires=">=2.7",
    entry_points='''
        [console_scripts]
        binaryai=binaryai.binaryai_cli:main
    '''
)



from .graphql.function import q_create_function, q_query_function, q_create_function_set, q_query_function_set, q_search_func_similarity
from .client import Client
from .error import BinaryAIException


def upload_function(
        client,
        name,
        feature,
        source_code=None,
        source_file=None,
        source_line=None,
        language=None,
        funcset_id=None
):
    '''
    upload function to BinaryAI server

    Args:
        client(binaryai.client.Client): Client instance
        name(string): name of the function
        feature(string): Feature of the function. Genertaed by feature extraction library. Encoding in base64.
        source_code(string): Source code of the function
        source_file(string): Source file of the function
        source_line(int): line number of the function
        language(string): Programming language of the function
        funcset_id(string): If functionSetID specified, it would be added into that set when adding function; if not, it would not be added into any set

    Returns:
        * **id** (string) -- id of this function
    '''
    if not isinstance(client, Client):
        raise BinaryAIException("SDK_ERROR", "Invalid client argument", None, None)
    var = {
        'name': name,
        'feature': feature,
        'sourceCode': source_code,
        'sourceFile': source_file,
        'sourceLine': source_line,
        'language': language,
        'functionSetId': funcset_id
    }
    r = client.execute(q_create_function, var)
    return r['createFunction']['function']['id']


def query_function(client, function_id):
    '''
    get function information by id

    Args:
        client(binaryai.client.Client): Client instance
        function_id(string): ID of the function

    Returns:
        * **function** (dict) -- Function's information
    '''
    if not isinstance(client, Client):
        raise BinaryAIException("SDK_ERROR", "Invalid client argument", None, None)
    var = {
        'funcId': function_id
    }
    r = client.execute(q_query_function, var)
    if r['function'] and function_id != r['function']['id']:
        raise BinaryAIException("SDK_ERROR", "Response function id not equal to the function_id", r, None)
    return r['function']


def create_function_set(client, function_ids=None):
    '''
    Create a new function set and add functions if needed

    Args:
        client(binaryai.client.Client): Client instance
        function_ids(list): Functions to be inserted into the new function set. Can be null so no functions will be added into the set.

    Returns:
        * **id** (string) -- id of the function set
    '''
    if not isinstance(client, Client):
        raise BinaryAIException("SDK_ERROR", "Invalid client argument", None, None)
    var = {
        'functionIds': function_ids
    }
    r = client.execute(q_create_function_set, var)
    return r['createFunctionSet']['functionSet']['id']


def query_function_set(client, funcset_id):
    '''
    get function set information by id

    Args:
        client(binaryai.client.Client): Client instance
        funcset_id(string): id of the function set

    Returns:
        * **functionSet** (dict) -- functionSet's information
    '''
    if not isinstance(client, Client):
        raise BinaryAIException("SDK_ERROR", "Invalid client argument", None, None)
    var = {
        'funcSetId': funcset_id
    }
    r = client.execute(q_query_function_set, var)
    if funcset_id != r['functionSet']['id']:
        raise BinaryAIException("SDK_ERROR", "Response function set id not equal to the funcset_id", r, None)
    return r['functionSet']


def search_sim_funcs(client, function_id, funcset_ids=None, topk=1):
    '''
    search top similar functions of the function

    Args:
        client(binaryai.client.Client): Client instance
        function_id(string): id of the function
        funcset_ids(list): ids of the function set to be compared, None means BinaryAI official sets.
        topk(int): return first topk results, default value is 1.

    Returns:
        * **similarity** (list): list of the top similarity functions
    '''
    if not isinstance(client, Client):
        raise BinaryAIException("SDK_ERROR", "Invalid client argument", None, None)
    var = {
        'funcId': function_id,
        'setId': funcset_ids,
        'topk': topk
    }
    r = client.execute(q_search_func_similarity, var)
    return r['function']['similarity']



# coding: utf-8
import idaapi
import idautils
import binaryai as bai


BINAYRAI_TOKEN = "PLEASE INPUT YOUR TOKEN HERE"
BINAYRAI_URL = "https://api.binaryai.tencent.com/v1/endpoint"


def upload_binary(client, funset_id, minsize=3):
    for func_ea in idautils.Functions():
        pfn = idaapi.get_func(func_ea)
        func_name = idaapi.get_func_name(func_ea)
        if idaapi.FlowChart(pfn).size < minsize:
            continue
        func_feat = bai.ida.get_func_feature(func_ea)
        if func_feat is None:
            continue
        bai.function.upload_function(client, func_name, func_feat, funcset_id=funset_id)


def main():
    client = bai.client.Client(BINAYRAI_TOKEN, BINAYRAI_URL)
    funset_id = bai.function.create_function_set(client)
    upload_binary(client, funset_id)
    print("[BinaryAI] BINARY_FUNCTIONSET_ID:", funset_id)


if __name__ == "__main__":
    main()
